using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Xml;

namespace System.Xml
{
  public class Xexpression
  {
    private static int debug_depth = 0;

    public static int DebugDepth {
      get { return debug_depth; }
      set { debug_depth = value; }
    }

    public struct Symbol : IEquatable<Symbol>
    {
      private static NameTable nt = new NameTable ();

      private string name;

      public Symbol (string str) { name = nt.Add (str); }

      public static implicit operator Symbol (string str)
      {
	return new Symbol (str);
      }

      public static implicit operator string (Symbol sym) { return sym.name; }

      public static bool operator== (Symbol n1, Symbol n2)
	{
	  return (object) n1.name == (object) n2.name;
	}

      public static bool operator!= (Symbol n1, Symbol n2)
	{
	  return (object) n1.name != (object) n2.name;
	}

      public static bool operator== (Symbol n1, string n2)
	{
	  return (object) n1.name == (object) n2;
	}

      public static bool operator!= (Symbol n1, string n2)
	{
	  return (object) n1.name != (object) n2;
	}

      public static bool operator== (string n1, Symbol n2)
	{
	  return (object) n1 == (object) n2.name;
	}

      public static bool operator!= (string n1, Symbol n2)
	{
	  return (object) n1 != (object) n2.name;
	}

      public bool Equals (Symbol name)
      {
	return Object.ReferenceEquals (this.name, name.name);
      }

      public override bool Equals (object obj)
      {
	return Object.ReferenceEquals (this.name, obj);
      }

      public override int GetHashCode ()
      {
	return name.GetHashCode ();
      }

      public static NameTable NameTable { get { return nt; } }

      public override string ToString () { return name; }
    }

    private static Symbol Qexpr = "expr";

    private static Symbol Qnull = "";
    private static Symbol Qfuncall = "funcall";
    private static Symbol Qinteger = "integer";
    private static Symbol Qstring = "string";
    private static Symbol Qsymbol = "symbol";
    private static Symbol Qlist = "list";

    public static Symbol Qdefun = "defun";
    public static Symbol Qdefmacro = "defmacro";
    private static Symbol Qfname = "fname";
    private static Symbol Qargs = "args";
    private static Symbol Qfixed = "fixed";
    private static Symbol Qoptional = "optional";
    private static Symbol Qrest = "rest";

    public static Symbol Qdefvar = "defvar";
    private static Symbol Qvname = "vname";
    private static Symbol Qdescription = "description";
    private static Symbol Qrange = "range";

    public abstract class Function
    {
      public Symbol name;
      public int min_arg, max_arg;

      public Function () { }

      public Function (Symbol name, int min_arg, int max_arg)
      {
	this.name = name;
	this.min_arg = min_arg;
	this.max_arg = max_arg;
      }

      public abstract Term Call (Domain domain, Variable vari, Term[] args);

      public override string ToString ()
      {
	return name;
      }

      internal class Subroutine : Function
      {
	public Builtin builtin;
	public bool setvar;

	public Subroutine (Builtin builtin, Symbol name, bool setvar,
			   int min_arg, int max_arg)
	  : base (name, min_arg, max_arg)
	  {
	    this.builtin = builtin;
	    this.setvar = setvar;
	  }

	public override Term Call (Domain domain, Variable vari, Term[] args)
	{
	  if (args != null)
	    {
	      args = (Term[]) args.Clone ();
	      for (int i = 0; i < args.Length; i++)
		{
		  args[i] = args[i].Eval (domain);
		  if (domain.Thrown)
		    return args[i];
		}
	    }
	  return builtin (domain, vari, args);
	}
      }

      internal class SpecialForm : Function
      {
	public Builtin builtin;

	public SpecialForm (Builtin builtin, Symbol name,
			    int min_arg, int max_arg)
	  : base (name, min_arg, max_arg)
	  {
	    this.builtin = builtin;
	  }

	public override Term Call (Domain domain, Variable vari, Term[] args)
	{
	  return builtin (domain, vari, args);
	}
      }

      private static void parse_head (Domain domain, XmlNode node,
				      out Symbol name,
				      out int min_arg, out int max_arg,
				      out Variable[] args)
      {
	int nfixed = 0;
	int noptional = 0;
	int nrest = 0;
	name = node.Attributes[Qfname].Value;
	    
	node = node.FirstChild;
	if (node != null && node.Name == Qargs)
	  {
	    XmlNode n;
	    for (n = node.FirstChild; n != null; n = n.NextSibling)
	      {
		if (n.Name == Qfixed)
		  nfixed++;
		else if (n.Name == Qoptional)
		  noptional++;
		else if (n.Name == Qrest)
		  nrest++;
		else
		  throw new Exception ("Invalid argument type: " + n);
	      }
	    min_arg = nfixed;
	    max_arg = nfixed + noptional + nrest;
	    args = new Variable[max_arg];
	    n = node.FirstChild;
	    for (int i = 0; i < max_arg; n = n.NextSibling)
	      args[i++] = domain.Defvar ((Symbol) n.Attributes[0].Value);
	    if (nrest == 1)
	      max_arg = - max_arg;
	  }
	else
	  {
	    min_arg = max_arg = 0;
	    args = null;
	  }
      }

      private static void parse_body (Domain domain, XmlNode node,
				      out Term[] body)
      {
	for (node = node.FirstChild; node != null; node = node.NextSibling)
	  if (node.Name != Qdescription
	      && node.Name != Qargs)
	    break;
	int nterms = 0;
	for (XmlNode n = node; n != null; n = n.NextSibling)
	  nterms++;
	if (nterms > 0)
	  {
	    body = new Term[nterms];
	    for (nterms = 0; node != null; node = node.NextSibling, nterms++)
	      body[nterms] = new Term (domain, node);
	  }
	else
	  body = null;
      }

      internal class Macro : Function
      {
	internal Variable[] args;
	internal Term[] body;

	public Macro (Domain domain, XmlNode node)
	  {
	    parse_head (domain, node, out name, out min_arg, out max_arg,
			out args);
	  }

	public void SetBody (Domain domain, XmlNode node)
	{
	  parse_body (domain, node, out body);
	}

	public override Term Call (Domain domain, Variable vari, Term[] args)
	{
	  Bindings current = domain.bindings;
	  Term result = Zero;

	  try {
	    for (int i = 0; i < min_arg; i++)
	      domain.Bind (this.args[i], args[i]);
	    if (body != null)
	      {
		try {
		  domain.Catch (CatchTag.Return);
		  foreach (Term term in body)
		    {
		      result = term.Eval (domain);
		      if (domain.Thrown)
			return result;
		    }
		} finally {
		  domain.Uncatch ();
		}
	      }
	  } finally {
	    domain.UnboundTo (current);
	  }
	  return result;
	}
      }

      internal class Lambda : Function
      {
	internal Variable[] args;
	internal Term[] body;

	public Lambda (Domain domain, XmlNode node)
	  {
	    parse_head (domain, node, out name, out min_arg, out max_arg,
			out args);
	  }

	public Lambda (Domain domain, Symbol name, Symbol[] args)
	  {
	    int nfixed = 0;
	    int noptional = 0;
	    int nrest = 0;

	    this.name = name;
	    if (args != null)
	      {
		int i = 0;
		for (i = 0; i < args.Length; i++, nfixed++)
		  if (args[i] == Qoptional || args[i] == Qrest)
		    break;
		if (i < args.Length)
		  {
		    if (args[i] == Qoptional)
		      {
			for (i++; i < args.Length; i++, noptional++)
			  if (args[i] == Qrest)
			    break;
			if (i < args.Length)
			  nrest = 1;
		      }
		  }
		min_arg = nfixed;
		max_arg = nfixed + noptional + nrest;
		this.args = new Variable[max_arg];
		int j;
		for (i = j = 0; j < this.args.Length; i++)
		  if (args[i] != Qoptional || args[i] != Qrest)
		    this.args[j++] = domain.Defvar (args[i]);
	      }
	    else
	      {
		min_arg = max_arg = 0;
	      }
	  }

	public void SetBody (Domain domain, XmlNode node)
	{
	  parse_body (domain, node, out body);
	}

	public void SetBody (Term[] body)
	{
	  this.body = body;
	}

	public override Term Call (Domain domain, Variable vari, Term[] args)
	{
	  Bindings current = domain.bindings;
	  Term result = Zero;

	  try {
	    int i;
	    if (args != null)
	      {
		Term[] newargs = new Term[args.Length];
		for (i = 0; i < min_arg; i++)
		  newargs[i] = args[i].Eval (domain);
		args = newargs;
	      }
	    for (i = 0; i < min_arg; i++)
	      domain.Bind (this.args[i], args[i]);
	    if (body != null)
	      {
		try {
		  domain.Catch (CatchTag.Return);
		  foreach (Term term in body)
		    {
		      result = term.Eval (domain);
		      if (domain.Thrown)
			return result;
		    }
		} finally {
		  domain.Uncatch ();
		}
	      }
	  } finally {
	    domain.UnboundTo (current);
	  }
	  return result;
	}
      }
    }

    public class Variable
    {
      public Domain domain;
      public readonly Symbol name;
      protected Term val;

      public Variable (Domain domain, Symbol name, Term val)
      {
	this.domain = domain;
	this.name = name;
	this.val = val;
      }

      public virtual bool ValueP (Term val) { return true; }

      public virtual Variable Clone (Domain domain)
      {
	return new Variable (domain, name, val);
      }

      public virtual void Reset () { val = Zero; }

      public virtual string Description { get { return null; } }

      public Term Value
      {
	get { return val; }
	set {
	  if (! ValueP (value))
	    throw new Exception ("Invalid value: " + value);
	  val = value;
	}
      }

      public Term SetValue (int i)
      {
	val.intval = i;
	val.objval = null;
	return val;
      }

      public Term SetValue (string s)
      {
	val.objval = s;
	return val;
      }

      public override string ToString () { return name + "(" + val + ")"; }

      public abstract class Typed : Variable
      {
	protected string desc;
	private Term default_val;

	protected Typed (Domain domain, Symbol name, Term val, string desc)
	  : base (domain, name, val)
	  {
	    this.desc = desc;
	    default_val = val;
	  }

	public override void Reset () { val = default_val; }

	public override string Description { get { return desc; } }
      }

      public class Int : Typed
      {
	private int[] range;

	public bool IsSubrange (int[] r)
	  {
	    if (range == null)
	      return true;
	    for (int i = 0; i < r.Length; i += 2)
	      {
		int j;
		for (j = 0; j < range.Length; j += 2)
		  if (range[j] <= r[i] && range[j + 1] >= r[i + 1])
		    break;
		if (j >= range.Length)
		  return false;
	      }
	    return true;
	  }

	public Int (Domain domain, Symbol name, int n, string desc, int[] range)
	  : base (domain, name, new Term (n), desc)
	  {
	    if (range != null && range.Length % 2 == 1)
	      throw_exception ("Range length for {0} not even", name);
	    this.range = range;
	    if (! ValueP (val))
	      throw_exception ("Invalid integer value for {0}: {1}", name, val);
	  }

	public override bool ValueP (Term term)
	{
	  if (! term.IsInt)
	    return false;
	  if (range == null)
	    return true;
	  int n = term.Intval;
	  for (int i = 0; i < range.Length; i += 2)
	    if (range[i] <= n && range[i + 1] >= n)
	      return true;
	  return false;
	}

	public override Variable Clone (Domain domain)
	{
	  return new Int (domain, name, val.Intval, desc, range);
	}

	public int[] Range { get { return range; } set { range = value; } }
      }

      public class Str : Typed
      {
	private string[] range;

	public bool IsSubrange (string[] r)
	  {
	    if (range == null)
	      return true;
	    for (int i = 0; i < r.Length; i++)
	      {
		int j;
		for (j = 0; j < range.Length; j++)
		  if (range[j] == r[i])
		    break;
		if (j >= range.Length)
		  return false;
	      }
	    return true;
	  }

	public Str (Domain domain, Symbol name, string str, string desc,
		    string[] range)
	  : base (domain, name, new Term (str), desc)
	  {
	    this.range = range;
	    if (! ValueP (val))
	      throw_exception ("Invalid string value for {0}: {1}", name, val);
	  }

	public override bool ValueP (Term term)
	{
	  if (! term.IsStr)
	    return false;
	  if (range == null)
	    return true;
	  string str = term.Strval;
	  foreach (string s in range)
	    if (str == s)
	      return true;
	  return false;
	}

	public override Variable Clone (Domain domain)
	{
	  return new Str (domain, name, val.Strval, desc, range);
	}

	public string[] Range { get { return range; } set { range = value; } }
      }

      public class Sym : Typed
      {
	public Symbol[] range;

	public bool IsSubrange (Symbol[] r)
	  {
	    if (range == null)
	      return true;
	    for (int i = 0; i < r.Length; i++)
	      {
		int j;
		for (j = 0; j < range.Length; j++)
		  if (range[j] == r[i])
		    break;
		if (j >= range.Length)
		  return false;
	      }
	    return true;
	  }

	public Sym (Domain domain, Symbol name, Symbol sym, string desc,
		    Symbol[] range)
	  : base (domain, name, new Term (sym), desc)
	  {
	    this.range = range;
	    if (! ValueP (val))
	      throw_exception ("Invalid symbol value for {0}: {1}", name, val);
	  }

	public override bool ValueP (Term term)
	{
	  if (! term.IsSymbol)
	    return false;
	  if (range == null)
	    return true;
	  Symbol name = term.Symval;
	  foreach (Symbol n in range)
	    if (name == n)
	      return true;
	  return false;
	}

	public override Variable Clone (Domain domain)
	{
	  return new Sym (domain, name, val.Symval, desc, range);
	}

	public Symbol[] Range { get { return range; } set { range = value; } }
      }
    }

    internal class Bindings
    {
      private Variable vari;
      private Term old_value;
      private Bindings next;
	
      private Bindings (Variable vari)
      {
	this.vari = vari;
	old_value = vari.Value;
      }
	
      public static Bindings Bind (Bindings bindings, Variable vari, Term val)
      {
	Bindings b = new Bindings (vari);

	b.vari.Value = val;
	b.next = bindings;
	return b;
      }

      internal Bindings UnboundTo (Bindings boundary)
      {
	for (Bindings b = this; b != boundary; b = b.next)
	  b.vari.Value = b.old_value;
	return boundary;
      }

      public override string ToString ()
      {
	string str = "(bindings";
	for (Bindings b = this; b != null; b = b.next)
	  str += " " + vari.name + "=" + b.old_value;
	return str + ")";
      }
    }

    private static void throw_exception (string msg)
    {
      throw new Exception (msg);
    }

    private static void throw_exception (string fmt, params object[] args)
    {
      throw new Exception (String.Format (fmt, args));
    }

    internal class CatchTag : IEquatable<CatchTag>
    {
      private object val;

      public CatchTag (Symbol sym) { val = (string) sym; }
      private CatchTag (int i) { val = i; }

      public static CatchTag Return = new CatchTag (0);
      public static CatchTag Break = new CatchTag (1);

      public static bool operator== (CatchTag t1, CatchTag t2)
	{ return t1.val == t2.val; }

      public static bool operator!= (CatchTag t1, CatchTag t2)
	{ return t1.val != t2.val; }

      public bool Equals (CatchTag tag) { return this.val == tag.val; }
      public override bool Equals (object val) { return this.val == val; }

      public override int GetHashCode () { return val.GetHashCode (); }
    }

    public class Domain
    {
      public Symbol name;
      public object context;
      public int depth = 0;

      internal Dictionary<Symbol, TermType> termtypes
	= new Dictionary<Symbol, TermType> ();
      internal Dictionary<Symbol, Function> functions
	= new Dictionary<Symbol, Function> ();
      internal Dictionary<Symbol, Variable> variables
	= new Dictionary<Symbol, Variable> ();
      internal Bindings bindings;
      private Stack<CatchTag> catch_stack = new Stack<CatchTag> ();
      private int catch_count = 0;

      internal Domain (Symbol name) { this.name = name; }

      public Domain (Symbol name, object context)
        : this (name, basic, context) { }

      public Domain (Symbol name, Domain parent, object context) : this (name)
      {
	termtypes = new Dictionary<Symbol, TermType> (parent.termtypes);
	functions = new Dictionary<Symbol, Function> (parent.functions);
	variables = new Dictionary<Symbol, Variable> (parent.variables);
	this.context = context;
      }

      internal void Bind (Variable vari, Term value)
      {
	bindings = Bindings.Bind (bindings, vari, value);
      }

      internal void UnboundTo (Bindings boundary)
      {
	if (bindings != null)
	  bindings = bindings.UnboundTo (boundary);
      }

      internal void Catch (CatchTag tag)
      {
	catch_stack.Push (tag);
	catch_count++;
      }

      internal void Uncatch ()
      {
	catch_stack.Pop ();
	if (catch_count > catch_stack.Count)
	  catch_count--;
      }

      public bool Thrown {
	get { return catch_count < catch_stack.Count; }
      }

      internal void ThrowReturn ()
      {
	foreach (CatchTag tag in catch_stack)
	  {
	    catch_count--;
	    if (tag == CatchTag.Return)
	      break;
	  }
      }

      internal void ThrowBreak ()
      {
	if (catch_stack.Peek () != CatchTag.Break)
	  throw_exception ("No outer loop to break");
	catch_count--;
      }

      internal void ThrowTag (CatchTag tag)
      {
	foreach (CatchTag elt in catch_stack)
	  {
	    catch_count--;
	    if (elt == tag)
	      break;
	  }
      }

      public void DefTerm (Symbol name, TermParser parser)
      {
	if (termtypes.ContainsKey (name)
	    || functions.ContainsKey (name))
	  throw new Exception ("already defined: " + name);
	termtypes[name] = new TermType (name, parser);
      }

      public void DefSubr (Builtin builtin, string str, bool setvar,
			   int min_arg, int max_arg, params string[] aliases)
      {
	Symbol name = str;
	if (termtypes.ContainsKey (name))
	  throw new Exception ("already defined as termtype: " + name);
	Function func = new Function.Subroutine (builtin, name, setvar,
						 min_arg, max_arg);
	functions[name] = func;
	foreach (string a in aliases)
	  functions[(Symbol) a] = func;
      }

      public void DefSpecial (Builtin builtin, string str,
			      int min_arg, int max_arg,
			      params string[] aliases)
      {
	Symbol name = str;
	if (termtypes.ContainsKey (name))
	  throw new Exception ("already defined as termtype: " + name);
	Function func = new Function.SpecialForm (builtin, name,
						  min_arg, max_arg);
	functions[name] = func;
	foreach (string a in aliases)
	  functions[(Symbol) a] = func;
      }

      public void DefAlias (string alias, string str)
      {
	functions[(Symbol) alias] = functions[(Symbol) str];
      }

      public void Defun (Symbol name, Symbol[] args, Term[] body,
			 bool prototype)
      {
	Function func;

	if (termtypes.ContainsKey (name))
	  throw new Exception ("already defined as termtype: " + name);
	if (prototype || ! functions.TryGetValue (name, out func))
	  {
	    func = new Function.Lambda (this, name, args);
	    functions[name] = func;
	  }
	if (! prototype)
	  ((Function.Lambda) func).SetBody (body);
      }

      public void Defun (XmlNode node, bool prototype)
      {
	Symbol name = node.Attributes[Qfname].Value;
	if (termtypes.ContainsKey (name))
	  throw new Exception ("already defined as termtype: " + name);
	Function func;

	if (prototype || ! functions.TryGetValue (name, out func))
	  {
	    func = new Function.Lambda (this, node);
	    functions[name] = func;
	  }
	if (! prototype)
	  ((Function.Lambda) func).SetBody (this, node);
      }

      public void Defun (Function func)
      {
	functions[func.name] = func;
      }

      public void Defmacro (XmlNode node, bool prototype)
      {
	Symbol name = node.Attributes[Qfname].Value;
	if (termtypes.ContainsKey (name))
	  throw new Exception ("already defined as termtype: " + name);
	Function func;

	if (prototype || ! functions.TryGetValue (name, out func))
	  {
	    func = new Function.Macro (this, node);
	    functions[name] = func;
	  }
	if (! prototype)
	  ((Function.Macro) func).SetBody (this, node);
      }

      public Variable Defvar (XmlNode node)
      {
	Symbol name = node.Attributes[0].Value;
	String desc;
	Variable vari;

	node = node.FirstChild;
	if (node != null && node.Name == Qdescription)
	  {
	    desc = node.InnerText;
	    node = node.NextSibling;
	  }
	else
	  desc = null;
	if (node != null)
	  {
	    Symbol type = node.Name;
	    XmlNodeList range_list = null;
	    int nranges = 0;
	    string val = node.InnerText;

	    node = node.NextSibling;
	    if (node != null)
	      {
		range_list = node.ChildNodes;
		nranges = range_list.Count;
	      }

	    if (type == Qinteger)
	      {
		int n = parse_integer (val);
		int[] range = null;
		if (range_list != null)
		  {
		    range = new int[nranges * 2];
		    for (int i = 0; i < nranges; i++)
		      {
			XmlNode nd = range_list[i];
			if (nd.Name == Qrange)
			  {
			    range[i * 2]
			      = parse_integer (nd.FirstChild.InnerText);
			    range[i * 2 + 1]
			      = parse_integer (nd.LastChild.InnerText);
			  }
			else
			  {
			    range[i * 2]
			      = range[i * 2 + 1]
			      = parse_integer (nd.FirstChild.InnerText);
			  }
		      }
		  }
		vari = DefvarInt (name, n, desc, range);
	      }
	    else if (type == Qstring)
	      {
		string[] range = null;
		if (range_list != null)
		  {
		    range = new string[nranges];
		    for (int i = 0; i < nranges; i++)
		      range[i] = range_list[i].FirstChild.InnerText;
		  }
		vari = DefvarStr (name, val, desc, range);
	      }
	    else if (type == Qsymbol)
	      {
		Symbol[] range = null;
		if (range_list != null)
		  {
		    range = new Symbol[nranges];
		    for (int i = 0; i < nranges; i++)
		      range[i] = range_list[i].FirstChild.InnerText;
		  }
		vari = DefvarSym (name, (Symbol) val, desc, range);
	      }
	    else
	      throw new Exception ("Unknown type: " + type);
	  }
	else
	  {
	    if (variables.TryGetValue (name, out vari))
	      vari = vari.Clone (this);
	    else
	      vari = new Variable (this, name, Zero);
	    variables[name] = vari;
	  }
	Console.WriteLine (vari);
	return vari;
      }

      public Variable Defvar (Variable vari)
      {
	vari = vari.Clone (this);
	variables[vari.name] = vari;
	return vari;
      }

      internal Variable Defvar (Symbol name)
      {
	Variable vari = new Variable (this, name, Zero);
	variables[name] = vari;
	return vari;
      }

      public Variable DefvarInt (Symbol name, int n, string desc, int[] range)
      {
	Variable vari;

	if (variables.TryGetValue (name, out vari))
	  {
	    Variable.Int intvari = vari as Variable.Int;
	    if (intvari == null)
	      throw new Exception ("Variable type mismatch: " + name);
	    if (range == null)
	      range = intvari.Range;
	    else if (! intvari.IsSubrange (range))
	      throw new Exception ("Variable range mismatch: " + name);
	    if (desc == null)
	      desc = vari.Description;
	  }
	vari = new Variable.Int (this, name, n, desc, range);
	variables[name] = vari;
	return vari;
      }

      public Variable DefvarStr (Symbol name, string str, string desc,
				 string[] range)
      {
	Variable vari;

	if (variables.TryGetValue (name, out vari))
	  {
	    Variable.Str strvari = vari as Variable.Str;
	    if (strvari == null)
	      throw new Exception ("Variable type mismatch: " + name);
	    if (range == null)
	      range = strvari.Range;
	    else if (! strvari.IsSubrange (range))
	      throw new Exception ("Variable range mismatch: " + name);
	    if (desc == null)
	      desc = vari.Description;
	  }
	vari = new Variable.Str (this, name, str, desc, range);
	variables[name] = vari;
	return vari;
      }

      public Variable DefvarSym (Symbol name, Symbol sym, string desc,
				 Symbol[] range)
      {
	Variable vari;

	if (variables.TryGetValue (name, out vari))
	  {
	    Variable.Sym symvari = vari as Variable.Sym;
	    if (symvari == null)
	      throw new Exception ("Variable type mismatch: " + name);
	    if (range == null)
	      range = symvari.Range;
	    else if (! symvari.IsSubrange (range))
	      throw new Exception ("Variable range mismatch: " + name);
	    if (desc == null)
	      desc = vari.Description;
	  }
	vari = new Variable.Sym (this, name, sym, desc, range);
	variables[name] = vari;
	return vari;
      }

      internal Function GetFunc (Symbol name)
      {
	Function func;

	if (! functions.TryGetValue (name, out func))
	  throw new Exception ("Unknown function: " + name);
	return func;
      }

      public bool CopyFunc (Domain domain, Symbol name)
      {
	Function func = GetFunc (name);

	domain.functions[name] = func;
	return true;
      }

      public void CopyFunc (Domain domain)
      {
	foreach (KeyValuePair<Symbol, Function> kv in functions)
	  domain.functions[kv.Key] = kv.Value;
      }

      public Variable GetVar (Symbol name, bool create)
      {
	Variable vari;

	if (! variables.TryGetValue (name, out vari))
	  {
	    if (! create)
	      return null;
	    variables[name] = vari = new Variable (this, name, Zero);
	  }
	return vari;
      }

      public override string ToString ()
      {
	string str = "<(functions";
	foreach (KeyValuePair<Symbol, Function> kv in functions)
	  str += " " + kv.Key;
	str += ") (variabls";
	foreach (KeyValuePair<Symbol, Variable> kv in variables)
	  str += " " + kv.Value;
	str += ")";
	if (bindings != null)
	  str += " " + bindings;
	if (context != null)
	  str += " (" + context + ")";
	str += ">";
	return str;
      }

      internal void DebugWrite (bool head, string fmt, params string[] arg)
      {
	if (debug_depth > depth)
	  {
	    if (head)
	      {
		Console.WriteLine ();
		for (int i = 0; i < depth; i++)
		  Console.Write (" ");
	      }
	    Console.Write (fmt, arg);
	  }
      }

      public object SaveValues ()
      {
	Dictionary<Variable,Term> values = new Dictionary<Variable,Term> ();

	foreach (KeyValuePair<Symbol,Variable> kv in variables)
	  values[kv.Value] = kv.Value.Value.Clone ();
	return values;
      }

      public void RestoreValues (object values)
      {
	foreach (KeyValuePair<Variable,Term> kv
		 in (Dictionary<Variable,Term>) values)
	  kv.Key.Value = kv.Value;
      }
    }

    public delegate Term Builtin (Domain domain, Variable vari, Term[] args);

    private static Domain basic = new Domain ("basic");

    static Xexpression ()
    {
      basic.DefTerm ("funcall", Funcall.parser);
      basic.DefTerm ("varref", Varref.parser);

      basic.DefSubr (Fset, "set", true, 1, 1, "=");
      basic.DefSubr (Fnot, "not", false, 1, 1, "!");
      basic.DefSubr (Fadd, "add", true, 1, -1, "+");
      basic.DefSubr (Fmul, "mul", true, 1, -1, "*");
      basic.DefSubr (Fsub, "sub", true, 1, -1, "-");
      basic.DefSubr (Fdiv, "div", true, 1, -1, "/");
      basic.DefSubr (Fmod, "mod", true, 1, 2, "%");
      basic.DefSubr (Flogior, "logior", true, 1, -1, "|");
      basic.DefSubr (Flogand, "logand", true, 1, -1, "&");
      basic.DefSubr (Flsh, "lsh", true, 1, 2, "<<");
      basic.DefSubr (Frsh, "rsh", true, 1, 2, ">>");
      basic.DefSubr (Feq, "eq", false, 2, -1, "==");
      basic.DefSubr (Fnoteq, "noteq", false, 2, 2, "!=");
      basic.DefSubr (Flt, "lt", false, 2, -1, "<");
      basic.DefSubr (Fle, "le", false, 2, -1, "<=");
      basic.DefSubr (Fgt, "gt", false, 2, -1, ">");
      basic.DefSubr (Fge, "ge", false, 2, -1, ">=");
      basic.DefSubr (Fappend, "append", true, 0, -1);
      basic.DefSubr (Fconcat, "concat", true, 0, -1);
      basic.DefSubr (Fnth, "nth", false, 2, 2);
      basic.DefSubr (Fcopy, "copy", false, 1, 1);
      basic.DefSubr (Fins, "ins", true, 2, 2);
      basic.DefSubr (Fdel, "del", true, 2, 2);
      basic.DefSubr (Feval, "eval", false, 1, 1);
      basic.DefSubr (Fbreak, "break", false, 0, 1);
      basic.DefSubr (Freturn, "return", false, 0, 1);

      basic.DefSpecial (Fand, "and", 1, -1, "&&");
      basic.DefSpecial (For, "or", 1, -1, "||");
      basic.DefSpecial (Fprogn, "progn", 0, -1, "expr");
      basic.DefSpecial (Fif, "if", 2, 3);
      basic.DefSpecial (Fwhen, "when", 1, -1);
      basic.DefSpecial (Floop, "loop", 1, -1);
      basic.DefSpecial (Fwhile, "while", 1, -1);
      basic.DefSpecial (Fcond, "cond", 1, -1);
      basic.DefSpecial (Fforeach, "foreach", 2, -1);
      basic.DefSpecial (Fquote, "quote", 1, 1);
      basic.DefSpecial (Ftype, "type", 1, 1);
      basic.DefSpecial (Fcatch, "catch", 2, 2);
      basic.DefSpecial (Fthrow, "throw", 1, 2);
    }

    private static Term Fset (Domain domain, Variable vari, Term[] args)
    {
      vari.Value = args[0];
      return args[0];
    }

    private static Term Fnot (Domain domain, Variable vari, Term[] args)
    {
      return args[0].IsTrue ? Zero : One;
    }

    private static Term Fadd (Domain domain, Variable vari, Term[] args)
    {
      int n = vari == null ? 0 : vari.Value.Intval;

      foreach (Term arg in args)
	n += arg.Intval;
      return (vari == null ? new Term (n) : vari.SetValue (n));
    }

    private static Term Fmul (Domain domain, Variable vari, Term[] args)
    {
      int n = vari == null ? 1 : vari.Value.Intval;
      foreach (Term arg in args)
	n *= arg.Intval;
      return (vari == null ? new Term (n) : vari.SetValue (n));
    }

    private static Term Fsub (Domain domain, Variable vari, Term[] args)
    {
      int n, i;

      if (vari == null)
	{
	  n = args[0].Intval;
	  i = 1;
	}
      else
	{
	  n = vari.Value.Intval;
	  i = 0;
	}
      while (i < args.Length)
	n -= args[i++].Intval;
      return (vari == null ? new Term (n) : vari.SetValue (n));
    }

    private static Term Fdiv (Domain domain, Variable vari, Term[] args)
    {
      int n, i;

      if (vari == null)
	{
	  n = args[0].Intval;
	  i = 1;
	}
      else
	{
	  n = vari.Value.Intval;
	  i = 0;
	}
      while (i < args.Length)
	n /= args[i++].Intval;
      return (vari == null ? new Term (n) : vari.SetValue (n));
    }

    private static Term Fmod (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval % args[1].Intval;

      return (vari == null ? new Term (n) : vari.SetValue (n));
    }

    private static Term Flogior (Domain domain, Variable vari, Term[] args)
    {
      int n = vari == null ? 0 : vari.Value.Intval;
      foreach (Term arg in args)
	n |= arg.Intval;
      return (vari == null ? new Term (n) : vari.SetValue (n));
    }

    private static Term Flogand (Domain domain, Variable vari, Term[] args)
    {
      int n, i;

      if (vari == null)
	{
	  n = args[0].Intval;
	  i = 1;
	}
      else
	{
	  n = vari.Value.Intval;
	  i = 0;
	}
      while (i < args.Length)
	n &= args[i++].Intval;
      return (vari == null ? new Term (n) : vari.SetValue (n));
    }

    private static Term Flsh (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval << args[1].Intval;
      return (vari == null ? new Term (n) : vari.SetValue (n));
    }

    private static Term Frsh (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval >> args[1].Intval;
      return (vari == null ? new Term (n) : vari.SetValue (n));
    }

    private static Term Feq (Domain domain, Variable vari, Term[] args)
    {
      Term o = args[0];

      if (o.objval == null)
	{
	  for (int i = 1; i < args.Length; i++)
	    if (args[i].objval != null || args[i].intval != o.intval)
	      return Zero;
	}
      else
	{
	  for (int i = 1; i < args.Length; i++)
	    if (o.objval.Equals (args[i].objval))
	      return Zero;
	}
      return One;
    }

    private static Term Fnoteq (Domain domain, Variable vari, Term[] args)
    {
      return Feq (domain, vari, args);
    }

    private static Term Flt (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval;

      for (int i = 1; i < args.Length; i++)
	{
	  int n1 = args[i].Intval;
	  if (n >= n1)
	    return Zero;
	  n = n1;
	}
      return One;
    }

    private static Term Fle (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval;
      for (int i = 1; i < args.Length; i++)
	{
	  int n1 = args[i].Intval;
	  if (n > n1)
	    return Zero;
	  n = n1;
	}
      return One;
    }

    private static Term Fgt (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval;
      for (int i = 1; i < args.Length; i++)
	{
	  int n1 = args[i].Intval;
	  if (n <= n1)
	    return Zero;
	  n = n1;
	}
      return One;
    }

    private static Term Fge (Domain domain, Variable vari, Term[] args)
    {
      int n = args[0].Intval;
      for (int i = 1; i < args.Length; i++)
	{
	  int n1 = args[i].Intval;
	  if (n < n1)
	    return Zero;
	  n = n1;
	}
      return One;
    }

    private static Term Fappend (Domain domain, Variable vari, Term[] args)
    {
      List<Term> list;

      if (vari == null)
	list = new List<Term> ();
      else
	list = vari.Value.Listval;

      foreach (Term arg in args)
	{
	  if (arg.IsList)
	    list.AddRange ((List<Term>) arg.objval);
	  else
	    list.Add (arg);
	}
      if (vari == null)
	{
	  Term result;
	  result.intval = 0;
	  result.objval = list;
	  return result;
	}
      return vari.Value;
    }

    private static Term Fconcat (Domain domain, Variable vari, Term[] args)
    {
      string str;

      if (vari == null)
	str = "";
      else
	str = vari.Value.Strval;

      foreach (Term arg in args)
	{
	  if (arg.IsStr)
	    str += (string) arg.objval;
	  else if (arg.IsList)
	    foreach (Term term in (List<Term>) arg.objval)
	      str += (char) term.Intval;
	  else
	    str += (char) arg.Intval;
	}

      if (vari == null)
	{
	  Term term;
	  term.intval = 0;
	  term.objval = str;
	  return term;
	}
      return vari.SetValue (str);
    }

    private static Term Fnth (Domain domain, Variable vari, Term[] args)
    {
      Term result;

      if (args[1].IsStr)
	{
	  result.intval = ((string) args[1].objval)[args[0].Intval];
	  result.objval = null;
	}
      else if (args[1].IsList)
	{
	  result = ((List<Term>) args[1].objval)[args[0].Intval];
	}
      else
	throw new Exception ("Term is not enumelable: " + args[1]);
      return result;
    }

    private static Term Fcopy (Domain domain, Variable vari, Term[] args)
    {
      Term result;

      result.intval = 0;
      result.objval = new List<Term> (args[0].Listval);
      return result;
    }

    private static Term Fins (Domain domain, Variable vari, Term[] args)
    {
      Term term = vari.Value;

      if (term.IsStr)
	{
	  string str = term.Strval.Insert (args[0].Intval, args[1].Strval);
	  vari.SetValue (str);
	}
      else if (vari.Value.IsList)
	vari.Value.Listval.InsertRange (args[0].Intval, args[1].Listval);
      else
	throw new Exception ("term is not collection: " + vari.Value);
      return vari.Value;
    }

    private static Term Fdel (Domain domain, Variable vari, Term[] args)
    {
      if (vari.Value.IsStr)
	{
	  string str
	    = vari.Value.Strval.Remove (args[0].Intval,
					args[1].Intval - args[0].Intval);
	  vari.SetValue (str);
	}
      else if (vari.Value.IsList)
	vari.Value.Listval.RemoveRange (args[0].Intval,
				      args[1].Intval - args[0].Intval);
      else
	throw new Exception ("term is not collection: " + vari.Value);
      return vari.Value;
    }

    private static Term Fand (Domain domain, Variable vari, Term[] args)
    {
      foreach (Term arg in args)
	if (! arg.Eval (domain).IsTrue)
	  return Zero;
      return One;
    }

    private static Term For (Domain domain, Variable vari, Term[] args)
    {
      foreach (Term arg in args)
	if (arg.Eval (domain).IsTrue)
	  return One;
      return Zero;
    }

    private static Term Feval (Domain domain, Variable vari, Term[] args)
    {
      return (args[0].Eval (domain));
    }

    private static Term Fprogn (Domain domain, Variable vari, Term[] args)
    {
      Term result = One;

      foreach (Term arg in args)
	result = arg.Eval (domain);
      return result;
    }

    private static Term Fif (Domain domain, Variable vari, Term[] args)
    {
      if (args[0].Eval (domain).IsTrue)
	return args[1].Eval (domain);
      if (args.Length == 2)
	return Zero;
      return args[2].Eval (domain);
    }

    private static Term Fwhen (Domain domain, Variable vari, Term[] args)
    {
      if (! args[0].Eval (domain).IsTrue)
	return Zero;
      Term result = One;
      for (int i = 1; i < args.Length; i++)
	result = args[i].Eval (domain);
      return result;
    }

    private static Term Freturn (Domain domain, Variable vari, Term[] args)
    {
      domain.ThrowReturn ();
      return args.Length == 0 ? Zero : args[0];
    }	  

    private static Term Fbreak (Domain domain, Variable vari, Term[] args)
    {
      domain.ThrowBreak ();
      return args.Length == 0 ? Zero : args[0];
    }	  

    private static Term Floop (Domain domain, Variable vari, Term[] args)
    {
      Term result = Zero;
      try {
	domain.Catch (CatchTag.Break);
	while (! domain.Thrown)
	  foreach (Term arg in args)
	    {
	      result = arg.Eval (domain);
	      if (domain.Thrown)
		return result;
	    }
      } finally {
	domain.Uncatch ();
      }
      return result;
    }

    private static Term Fwhile (Domain domain, Variable vari, Term[] args)
    {
      Term result = Zero;
      try {
	domain.Catch (CatchTag.Break);
	while (! domain.Thrown && args[0].Eval (domain).IsTrue)
	  for (int i = 1; i < args.Length; i++)
	    {
	      result = args[i].Eval (domain);
	      if (domain.Thrown)
		return result;
	    }
      } finally {
	domain.Uncatch ();
      }
      return result;
    }

    private static Term Fcond (Domain domain, Variable vari, Term[] args)
    {
      foreach (Term arg in args)
	{
	  List<Term> list = arg.Listval;
	  Term result = list[0].Eval (domain);

	  if (result.IsTrue)
	    {
	      for (int i = 1; i < list.Count; i++)
		result = list[i].Eval (domain);
	      return result;
	    }
	}
      return Zero;
    }

    private static Term Fforeach (Domain domain, Variable vari,
					Term[] args)
    {
      Term result = args[0].Eval (domain);
      if (domain.Thrown)
	return result;
      List<Term> list = result.Listval;
      Bindings current = domain.bindings;

      try {
	domain.Catch (CatchTag.Break);
	foreach (Term term in list)
	  {
	    domain.Bind (vari, term);
	    try {
	      for (int i = 1; i < args.Length; i++)
		{
		  result = args[i].Eval (domain);
		  if (domain.Thrown)
		    return result;
		}
	    } finally {
	      domain.UnboundTo (current);
	    }
	  }
      } finally {
	domain.Uncatch ();
      }
      return result;
    }

    private static Term Fquote (Domain domain, Variable vari, Term[] args)
    {
      return new Term (args[0]);
    }

    private static Term Ftype (Domain domain, Variable vari, Term[] args)
    {
      if (args[0].IsInt)
	return TermInt;
      if (args[0].IsStr)
	return TermStr;
      if (args[0].IsSymbol)
	return TermSymbol;
      if (args[0].IsList)
	return TermList;
      return TermTerm;
    }

    public static Term Fcatch (Domain domain, Variable vari, Term[] args)
    {
      Term result = Zero;
      try {
	domain.Catch (new CatchTag (args[0].Symval));
	result = args[1].Eval (domain);
      } finally {
	domain.Uncatch ();
      }
      return result;
    }

    public static Term Fthrow (Domain domain, Variable vari, Term[] args)
    {
      domain.ThrowTag (new CatchTag (args[0].Symval));
      return (args.Length == 1 ? Zero : args[1]);
    }

    public delegate TermValue TermParser (Domain domain, XmlNode node);

    public class TermType
    {
      public readonly Symbol type;
      internal readonly TermParser parser;

      public TermType (Symbol type, TermParser parser)
      {
	this.type = type;
	this.parser = parser;
      }
    }

    public abstract class TermValue
    {
      public virtual Term Eval (Domain domain) { return new Term (this); }
      public virtual TermValue Clone () { return this; }
    }

    private class Varref : TermValue
    {
      private Symbol name;
      private Variable vari;

      public Varref (Symbol name) { this.name = name; }

      public override Term Eval (Domain domain)
      {
	if (vari == null || vari.domain != domain)
	  vari = domain.GetVar (name, true);
	return vari.Value;
      }

      internal static TermValue parser (Domain domain, XmlNode node)
      {
	return new Varref ((Symbol) node.Attributes[Qvname].Value);
      }

      public override string ToString ()
      {
	return "<varref vname=\"" + name + "\"/>";
      }
    }

    private class Funcall : TermValue
    {
      internal Function func;
      internal Variable vari;
      internal Term[] args;

      public Funcall (Function func, Variable vari, Term[] args)
	{
	  if (args != null)
	    {
	      int nargs = args.Length;
	      if (nargs < func.min_arg
		  || (func.max_arg >= 0 && nargs > func.max_arg))
		throw_exception ("Invalid number of arguments to {0}: ",
				 func.name, nargs);
	    }
	  this.func = func;
	  this.vari = vari;
	  this.args = args;
	}

      internal static TermValue parser (Domain domain, XmlNode node)
	{
	  Symbol fname = node.Name;
	  XmlAttribute attr;

	  if (fname == Qfuncall)
	    fname = node.Attributes[Qfname].Value;
	  Function func = domain.GetFunc (fname);
	  Variable vari;
	  attr = node.Attributes[Qvname];
	  vari = attr == null ? null : domain.GetVar (attr.Value, true);
	  XmlNodeList nlist = node.ChildNodes;
	  int nargs = nlist.Count;
	  Term[] args = new Term[nargs];
	  for (int i = 0; i < nargs; i++)
	    args[i] = new Term (domain, nlist[i]);
	  return new Funcall (func, vari, args);
	}

      public override Term Eval (Domain domain)
      {
	domain.DebugWrite (true, ToString ());
	domain.depth++;
	Term result = func.Call (domain, vari, args);
	domain.depth--;
	domain.DebugWrite (true, " ==> {0}", result.ToString ());
	return result;
      }

      public override TermValue Clone ()
      {
	return new Funcall (func, vari, args);
      }

      public override string ToString ()
      {
	string str = "<" + func.name;
	if (vari != null)
	  str += " vname=\"" + vari.name + "\"";
	if (args == null)
	  return str + "/>";
	str += ">";
	if (func is Function.SpecialForm)
	  str += String.Format ("({0})...", args.Length);
	else
	  foreach (Term e in args)
	    str += e;
	return (str + "</" + func.name + ">");
      }
    }

    public struct Term
    {
      public int intval;
      public object objval;

      // <integer>...</integer>
      public Term (int i) { intval = i; objval = null; }
      // <symbol>...</symbol>
      public Term (Symbol name) { intval = 0; objval = name; }
      // <string>...</string>
      public Term (string str) { intval = 0; objval = str; }
      // <list>...</list>
      public Term (List<Term> list) { intval = 0; objval = list; }

      public Term (Term term) { intval = term.intval; objval = term.objval; }
      public Term (TermValue obj) { intval = 0; objval = obj; }

      public Term (Domain domain, XmlNode node)
	{
	  Symbol name = node.Name;

	  if (name == Qinteger)
	    {
	      intval = parse_integer (node.InnerText);
	      objval = null;
	    }
	  else
	    {
	      intval = 0;
	      if (name == Qsymbol)
		objval = (Symbol) node.InnerText;
	      else if (name == Qstring)
		objval = node.InnerText.Clone ();
	      else if (name == Qlist)
		{
		  List<Term> list = new List<Term> ();
		  for (node = node.FirstChild; node != null;
		       node = node.NextSibling)
		    list.Add (new Term (domain, node));
		  objval = list;
		}
	      else
		{
		  TermType term_type;

		  if (domain.termtypes.TryGetValue (name, out term_type))
		    objval = term_type.parser (domain, node);
		  else
		    {
		      Funcall funcall = (Funcall) Funcall.parser (domain, node);
		      if (funcall.func is Function.Macro)
			{
			  Term result = funcall.Eval (domain);
			  intval = result.intval;
			  objval = result.objval;
			}
		      else
			objval = funcall;
		    }
		}
	    }
	}

      // <varref vname="VNAME"/>
      public Term (Domain domain, Symbol vname)
	{
	  intval = 0;
	  objval = new Varref (vname);
	}

      // <funcall fname="FNAME">...</funcall>
      public Term (Domain domain, Symbol fname, Term[] args)
	: this (domain, fname, Qnull, args) { }

      // <funcall fname="FNAME" vname="VNAME">...</funcall>
      public Term (Domain domain, Symbol fname, Symbol vname, Term[] args)
	{
	  intval = 0;

	  Function func = domain.GetFunc (fname);
	  Variable vari = vname == Qnull ? null : domain.GetVar (vname, true);
	  Funcall funcall = new Funcall (func, vari, args);
	  if (func is Function.Macro)
	    {
	      Term result = funcall.Eval (domain);
	      intval = result.intval;
	      objval = result.objval;
	    }
	  else
	    objval = funcall;
	}

      public object Objval {
	get {
	  if (objval == null)
	    throw new Exception ("term is an integer: " + this);
	  return objval;
	}
      }

      public int Intval {
	get {
	  if (objval != null)
	    throw new Exception ("term is not integer: " + this);
	  return intval;
	}
      }

      public string Strval {
	get {
	  if (! IsStr)
	    throw new Exception ("term is not string: " + this);
	  return (string) objval;
	}
      }

      public string Symval {
	get {
	  if (! IsSymbol)
	    throw new Exception ("term is not symbol: " + this);
	  return (Symbol) objval;
	}
      }

      public List<Term> Listval {
	get {
	  if (! IsList)
	    throw new Exception ("term is not list: " + this);
	  return (List<Term>) objval;
	}
      }

      public bool IsTrue {
	get {
	  return (objval == null
		  ? (intval != 0)
		  : objval is List<Term>
		  ? (((List<Term>) objval).Count != 0)
		  : true);
	}
      }
      public bool IsInt { get { return (objval == null); } }
      public bool IsStr { get { return (objval is string); } }
      public bool IsSymbol { get { return (objval is Symbol); } }
      public bool IsList { get { return (objval is List<Term>); } }

      public bool IsType (Type type)
      {
	return (objval == null ? type == typeof (int)
		: type == objval.GetType ());
      }

      public Term Eval (Domain domain)
      {
	if (objval == null || objval is Symbol || objval is string)
	  return this;
	if (objval is List<Term>)
	  return new Term ((List<Term>) objval);
	return ((TermValue) objval).Eval (domain);
      }

      public Term Clone ()
      {
	if (objval == null || objval is Symbol || objval is string)
	  return this;
	if (objval is List<Term>)
	  {
	    List<Term> list = new List<Term> ();
	    list.InsertRange (0, ((List<Term>) objval));
	    return new Term (list);
	  }
	return new Term (((TermValue) objval).Clone ());
      }

      public override string ToString ()
      {
	string str;

	if (objval == null)
	  str = "<integer>" + intval + "</integer>";
	else if (objval is Symbol)
	  str = "<symbol>" + objval + "</symbol>";
	else if (objval is string)
	  str = "<string>" + objval + "</string>";
	else if (objval is List<Term>)
	  {
	    str = "<list>";
	    foreach (Term e in (List<Term>) objval)
	      str += e;
	    str += "</list>";
	  }
	else if (objval is Term)
	  str = "<quote>" + objval + "</quote>";
	else if (objval is TermValue)
	  str = ((TermValue) objval).ToString ();
	else
	  throw new Exception ("invalid Term object: " + objval);
	return str;
      }
    }

    static private Term Zero = new Term (0);
    static private Term One = new Term (1);
    static private Term TermInt = new Term (Qinteger);
    static private Term TermStr = new Term (Qstring);
    static private Term TermSymbol = new Term (Qsymbol);
    static private Term TermList = new Term (Qlist);
    static private Term TermTerm = new Term ((Symbol) "term");

    internal static int parse_integer (string str)
    {
      int len = str.Length;
      bool negative = false;

      if (len <= 1)
	return (len == 0 ? 0 : str[0] - '0');

      int c = str[0];
      int i;

      if (c == '?')
	return str[1];
      if ((c == '0' || c == '#') && str[1] == 'x')
	{
	  i = 0;
	  for (int idx = 2; idx < len; idx++)
	    {
	      c = str[idx];
	      if (c < '0')
		break;
	      else if (c <= '9')
		i = i * 16 + (c - '0');
	      else if (c < 'A')
		break;
	      else if (c <= 'F')
		i = i * 16 + (c - 'A');
	      else if (c < 'a')
		break;
	      else if (c <= 'f')
		i = i * 16 + (c - 'a');
	      else
		break;
	    }
	  return i;
	}
      if (c == '-')
	negative = true;
      i = c - '0';
      for (int idx = 1; idx < len; idx++)
	{
	  c = str[idx];
	  if (c < '0' || c > '9')
	    break;
	  i = i * 10 + (c - '0');
	}
      return negative ? - i : i;
    }

    private Term[] terms;

    public static Term[] ParseTerms (Domain domain, XmlNode node)
    {
      int nterms = 0;
      for (XmlNode n = node; n != null; n = n.NextSibling)
	if (n.NodeType == XmlNodeType.Element)
	  {
	    if (n.Name == Qdefun)
	      domain.Defun (n, true);
	    else if (n.Name == Qdefmacro)
	      domain.Defmacro (n, true);
	    else if (n.Name == Qdefvar)
	      domain.Defvar (n);
	    else
	      nterms++;
	  }
      Term[] terms = new Term[nterms];
      int i = 0;
      for (XmlNode n = node; n != null; n = n.NextSibling)
	if (n.NodeType == XmlNodeType.Element)
	  {
	    if (n.Name == Qdefun)
	      domain.Defun (n, false);
	    else if (n.Name == Qdefmacro)
	      domain.Defmacro (n, false);
	    else if (n.Name != Qdefvar)
	      terms[i++]= new Term (domain, n);
	  }
      return terms;
    }

    public static Term Eval (Domain domain, Term[] terms)
    {
      Term result = new Term (0);
      foreach (Term term in terms)
	{
	  result = term;
	  if (result.Objval is Funcall)
	    while ((result = result.Eval (domain)).Objval is Funcall);
	}
      return result;
    }


    public Xexpression (Domain domain, XmlNode node)
    {
      terms = ParseTerms (domain, node);
    }

    public Xexpression (Domain domain, string url)
    {
      XmlDocument doc = new XmlDocument (Symbol.NameTable);
      XmlNode node;

      using (XmlTextReader reader = new XmlTextReader (url, doc.NameTable))
	{
	  do {
	    reader.Read ();
	  } while (reader.NodeType != XmlNodeType.None
		   && (reader.NodeType != XmlNodeType.Element
		       || reader.Name != Qexpr));
	  if (reader.NodeType == XmlNodeType.None)
	    throw new Exception ("Node <expr> not found");
	  node = doc.ReadNode (reader);
	}
      terms = ParseTerms (domain, node.FirstChild);
    }

    public Term Eval (Domain domain)
    {
      Term result = Zero;

      domain.depth = 0;
      try {
	domain.Catch (CatchTag.Return);
	foreach (Term term in terms)
	  {
	    result = term.Eval (domain);
	    if (domain.Thrown)
	      return result;
	  }
      } finally {
	domain.Uncatch ();
      }
      return result;
    }

    public override string ToString ()
    {
      string str = "";
      for (int i = 0; i < terms.Length; i++)
	str += terms[i];
      return str;
    }
  }
}
